import os

def _setup(foldername):
    """
    Sets up the script with the input and output directorys corresponding to my data organization strategy.
    
    Parameters
    ----------
    foldername : str
        name of directory containing raw data

    Returns
    -------
    input_dir : str
        path to input directory
    output_dir : str
        path to ouput directory
    """
    
    # script_dir = '/Volumes/GoogleDrive/My Drive/RG Lab/Microscopy/analysis/'
    script_dir = './'
    if os.path.exists(script_dir+foldername):
        pass
    else:
        os.mkdir(foldername)
    
    input_dir = '../'+foldername
    output_dir = script_dir+foldername+'/'
    
    return input_dir, output_dir


# Configuration file for the Sphinx documentation builder.
#
# This file only contains a selection of the most common options. For a full
# list see the documentation:
# https://www.sphinx-doc.org/en/master/usage/configuration.html

# -- Path setup --------------------------------------------------------------

import inspect
# If extensions (or modules to document with autodoc) are in another directory,
# add these directories to sys.path here. If the directory is relative to the
# documentation root, use os.path.abspath to make it absolute, like shown here.
#
import os
import shutil
import subprocess
import sys

try:
    from fraplib import __version__ as release
except ImportError:
    release = "unknown"



# -- Project information -----------------------------------------------------

project = "fraplib"
copyright = "2022, Elizabeth May"
author = "Elizabeth May"


# -- Generate API ------------------------------------------------------------
api_folder_name = "api"
shutil.rmtree(api_folder_name, ignore_errors=True)  # in case of new or renamed modules
subprocess.call(
    " ".join(
        [
            "sphinx-apidoc",
            f"-o {api_folder_name}/",
            "--force",
            "--no-toc",
            "--templatedir _templates",
            "--separate",
            "../fraplib/",
            # excluded modules
            # nothing here for cookiecutter
        ]
    ),
    shell=True,
)


# -- General configuration ---------------------------------------------------

# Add any Sphinx extension module names here, as strings. They can be
# extensions coming with Sphinx (named 'sphinx.ext.*') or your custom
# ones.
extensions = [
    "jupyter_sphinx",
    "myst_nb",
    "numpydoc",
    "sphinx.ext.autodoc",
    "sphinx.ext.intersphinx",
    "sphinx.ext.linkcode",
    "sphinx.ext.mathjax",
    "sphinx.ext.napoleon",
    "sphinx_copybutton",
    "sphinx_panels",
    "sphinx_thebe",
    "sphinx_togglebutton",
]


# API settings
autodoc_default_options = {
    "members": True,
    "show-inheritance": True,
    "undoc-members": True,
}
add_module_names = False
napoleon_google_docstring = False
napoleon_include_private_with_doc = False
napoleon_include_special_with_doc = False
napoleon_numpy_docstring = True
napoleon_use_admonition_for_examples = False
napoleon_use_admonition_for_notes = False
napoleon_use_admonition_for_references = False
napoleon_use_ivar = False
napoleon_use_param = False
napoleon_use_rtype = False
numpydoc_show_class_members = False

# Cross-referencing configuration
default_role = "py:obj"
primary_domain = "py"
nitpicky = True  # warn if cross-references are missing

# Intersphinx settings
intersphinx_mapping = {
    "ipywidgets": ("https://ipywidgets.readthedocs.io/en/stable", None),
    "matplotlib": ("https://matplotlib.org/stable", None),
    "numpy": ("https://numpy.org/doc/stable", None),
    "python": ("https://docs.python.org/3", None),
}

# remove panels css to get wider main content
panels_add_bootstrap_css = False

# Settings for copybutton
copybutton_prompt_is_regexp = True
copybutton_prompt_text = r">>> |\.\.\. "  # doctest

# Settings for linkcheck
linkcheck_anchors = False
linkcheck_ignore = []  # type: ignore

execution_timeout = -1
jupyter_execute_notebooks = "off"
if "EXECUTE_NB" in os.environ:
    print("\033[93;1mWill run Jupyter notebooks!\033[0m")
    jupyter_execute_notebooks = "force"

# Settings for myst-parser
myst_enable_extensions = [
    "amsmath",
    "colon_fence",
    "dollarmath",
    "smartquotes",
    "substitution",
]
suppress_warnings = [
    "myst.header",
]

# Add any paths that contain templates here, relative to this directory.
templates_path = ["_templates"]

# List of patterns, relative to source directory, that match files and
# directories to ignore when looking for source files.
# This pattern also affects html_static_path and html_extra_path.
exclude_patterns = [
    "**ipynb_checkpoints",
    ".DS_Store",
    "Thumbs.db",
    "_build",
]


# -- Options for HTML output -------------------------------------------------

# The theme to use for HTML and HTML Help pages.  See the documentation for
# a list of builtin themes.
#
html_copy_source = True  # needed for download notebook button
html_css_files = [
    "custom.css",
]
html_sourcelink_suffix = ""
html_static_path = ["_static"]
html_theme = "sphinx_book_theme"
html_theme_options = {
    "launch_buttons": {
        "binderhub_url": "https://mybinder.org",
        "colab_url": "https://colab.research.google.com",
        "notebook_interface": "jupyterlab",
        "thebe": True,
        "thebelab": True,
    },
    "path_to_docs": "docs",
    "repository_branch": "main",
    "repository_url": "https://github.com/emay2022/fraplib",
    "use_download_button": True,
    "use_edit_page_button": True,
    "use_issues_button": True,
    "use_repository_button": True,
}
html_title = "fraplib"

master_doc = "index"
thebe_config = {
    "repository_url": html_theme_options["repository_url"],
    "repository_branch": html_theme_options["repository_branch"],
}


# based on pandas/doc/source/conf.py
def linkcode_resolve(domain, info):
    """
    Determine the URL corresponding to Python object
    """
    if domain != "py":
        return None

    modname = info["module"]
    fullname = info["fullname"]

    submod = sys.modules.get(modname)
    if submod is None:
        return None

    obj = submod
    for part in fullname.split("."):
        try:
            obj = getattr(obj, part)
        except AttributeError:
            return None

    try:
        fn = inspect.getsourcefile(inspect.unwrap(obj))
    except TypeError:
        fn = None
    if not fn:
        return None

    try:
        source, lineno = inspect.getsourcelines(obj)
    except OSError:
        lineno = None

    if lineno:
        linespec = f"#L{lineno}-L{lineno + len(source) - 1}"
    else:
        linespec = ""

    fn = os.path.relpath(fn, start=os.path.dirname("../fraplib"))

    return f"https://github.com/emay2022/fraplib/blob/main/fraplib/{fn}{linespec}"  # noqa



import pandas as pd

def read_chromatogram(fpath):
    raw_data = pd.read_excel(fpath, header = 1)
    curves = raw_data.columns[::2].values
    curves_2x = raw_data.columns.values.copy()
    curves_2x[1::2] = curves
    curvelabels = [label.split('_')[-1] for label in curves_2x]
    title = raw_data.columns[::2].values[0].split(':')[0]
    units = [unit.strip() for unit in raw_data.iloc[0].values]
    cols = list(zip(*[curvelabels, units]))
    c = pd.MultiIndex.from_tuples(cols, names = ['curve', 'unit'])
    data = pd.DataFrame(data=raw_data.iloc[1:].values, columns = c)
    return data, curvelabels, title


import czifile
import xml.etree.ElementTree as et
import numpy as np


def version(file):
    """
    Get the CZI version from the metadata.
    
    Parameters
    ----------
    file : czifile
        original data file
    
    Returns
    -------
    version : str
    """
    md = file.metadata()
    mdt = et.fromstring(md)
    
    version = [elem.text for elem in mdt.findall('./Metadata/Version')]
    
    if version:
        version = version[0]
    else:
        version = [elem.get('Version') for elem in mdt.findall('./Metadata/Experiment')][0]
    
    return version


def channel_label(file):
    """
    Grabs the descriptive channel names from the metadata.

    Parameters
    ----------
    file : CziFile
        original data file

    Returns
    -------
    chlabels : list
        list of descriptive channel names (strings), in the same order as the channels in the image array.
    """

    md = et.fromstring(file.metadata())

    chlabels = []
    nTL = 0
    for ind, cm in enumerate(md.findall(".//Dimensions//ContrastMethod")):
        if cm.text == "Fluorescence":
            desc = md.findall(".//Dimensions//Fluor")[ind - nTL].text
            chlabels.append(desc)
        else:
            chlabels.append("Transmitted Light")
            nTL += 1
    
    if len(chlabels) == 1:
        chlabels = chlabels[0]
    
    return chlabels


def get_gain(file):
    """
    Grabs the gain info for each channel, in the same order as the channels in the image

    Parameters
    ----------
    file : CziFile
        original data file

    Returns
    -------
    gains : list
        list of gain values
    """

    v = version(file)
    md = et.fromstring(file.metadata())
    
    if v == '1.0':
        gains = [float(elem.text) for elem in md.findall(".//Dimensions//Gain")]
    else:
        channel = channel_label(file)
        gains = []
        for elem in md.findall('.//DetectorGain/../..'):
            # print(elem.tag, elem.attrib)
            a = elem.attrib
            if 'Name' in a:
                if isinstance(channel, str):
                    if channel in a['Name']:
                        # print(a['Name'])
                        gains.append(float(elem.findall('.//DetectorGain')[0].text))
                elif isinstance(channel, list):
                    for chname in channel:
                        if chname in a['Name']:
                            # print(a['Name'])
                            gains.append(float(elem.findall('.//DetectorGain')[0].text))
    
    if len(gains) == 1:
        gains = gains[0]
    
    return gains


def get_power(file):
    """
    Grabs the laser power info for each channel, in the same order as the channels in the image

    Parameters
    ----------
    file : CziFile
        original data file

    Returns
    -------
    powers : list
        list of power values; fractional
    """

    md = et.fromstring(file.metadata())
    powers = [
        round(1 - float(elem.text), 2)
        for elem in md.findall(".//LightSourceSettings/Attenuation")
    ]
    
    if len(powers) == 1:
        powers = powers[0]
    
    return powers


def get_objective(file):
    """
    Grabs the objectve info for the experiment

    Parameters
    ----------
    file : CziFile
        original data file

    Returns
    -------
    objective : str
        objective information
    """

    md = et.fromstring(file.metadata())
    v = version(file)
    if v == '1.0':
        objective = md.findall(".//Experiment//Objective")[0].text
    else:
        objective = [elem.attrib['Name'] for elem in md.findall('.//Objectives/Objective')][0]

    return objective


def get_em(file):
    """
    Grabs the emission wavelength range for each fluorescence channel in the experiment.

    Paramters
    ---------
    file : CziFile
        original data file

    Returns
    -------
    em : list of tuples
        (lower wavelength (nm), higher wavelength (nm))
    """

    md = et.fromstring(file.metadata())
    em = [
        (float(elem.text.split("-")[0]), float(elem.text.split("-")[1]))
        for elem in md.findall(".//Ranges")
    ]
    
    if len(em) == 1:
        em = em[0]
    
    return em


def get_ex(file):
    """
    Grabs the excitation wavelength (nm) for each channel in the experiment.

    Parameters
    ----------
    file : CziFile
        original data file

    Returns
    -------
    ex : list of floats
    """

    md = et.fromstring(file.metadata())
    ex = [float(elem.text) for elem in md.findall(".//LightSourceSettings/Wavelength")]
    
    if len(ex) == 1:
        ex = ex[0]

    return ex


def get_scales(file):
    """
    Grab the scaling info in all dimensions, in units of microns.
    
    Parameters
    ----------
    file : CziFile
        original data file
    
    Returns
    -------
    scales : list
        [x-scale, y-scale] or [x-scale, y-scale, z-scale], values in units of microns
    """

    md = et.fromstring(file.metadata())
    scales = {
        each.attrib["Id"] : float(each.findall("./Value")[0].text) / 1e-6
        for each in md.findall(".//Scaling/Items/Distance")
    }

    
    lst = list(zip( list(scales.keys()), list(scales.values() ))) # key : value pairs as list of tuples
    def get_first(tup):
        return tup[0]
    lst.sort(key = get_first) # sort lst alphabetically by the first element of each tuple (the key from above); produces [('X', value), ('Y', value), ('Z', value)]
    scales = [tup[1] for tup in lst] # list of values in X, Y, Z order
    
    return scales


def get_regions(file, units=None):
    """
    Grabs the geometry of regions of interest specified during the experiment. Currenltly works only for circles.

    Parameters
    ----------
    file : CziFile
        original data file
    units : str

    Returns
    -------
    geom : list of tuples
        e.g. (cx, cy, r)
    """

    md = et.fromstring(file.metadata())
    shapes = [elem.tag for elem in md.findall(".//Circle")]
    
    v = version(file)
    
    if v == '1.0':
        shape = shapes[::2]
    else:
        shape = shapes

    xum = get_scales(file)[0] # microns
    yum = get_scales(file)[1] # microns
        
    geom = []
    ind = 0
    for region in shape:
        if "Circle" in region:
            cx = float(md.findall(".//Circle//CenterX")[ind].text)
            cy = float(md.findall(".//Circle//CenterY")[ind].text)
            r = float(md.findall(".//Circle//Radius")[ind].text)
            
            if units == 'microns' or units == 'micrometers':
                
                cx = cx * xum
                cy = cy * yum
                r = r * xum
            
            geom.append((cx, cy, r))
            ind += 1 

    # if len(geom) == 1:
    #     geom = geom[0]
    
    return geom


def get_dims(file, specific = None):
    """
    Grabs the size of each dimension of the experiment.

    Parameters
    ----------
    file : CziFile
        original data file
    specific : str
        e.g. 'CZYX'

    Returns
    -------
    dims : list
    labels : list
        e.g. ['C', 'Z', 'Y', 'X']
    """

    md = et.fromstring(file.metadata())
    names = [letter for letter in file.axes]  # list
    shape = file.asarray().shape  # tuple

    dims_dict = {name: size for name, size in zip(names, shape)}
    
    if specific is None:
        dims = [ dims_dict[key] for key in dims_dict if dims_dict[key] > 1 ]
        labels = [ key for key in dims_dict if dims_dict[key] > 1 ]
    else:
        dims = [dims_dict[dim] for dim in specific]
        labels = specific

    return dims, labels


import fraplib
import pytest

from pathlib import Path
import czifile



@pytest.fixture()
def testfile():
    # put the data file to load in the test folder
    return czifile.CziFile(Path(__file__).parent / "testfile.czi")

def test_metadata(testfile):
    expected = {
        "SizeB": 1,
        "SizeV": 1,
        "SizeC": 1,
        "SizeT": 45,
        "SizeZ": 1,
        "SizeY": 256,
        "SizeX": 256,
        "Size0": 1,
    }
    assert expected == fraplib.get_dims(testfile)

    expected = [(638.0000000000002, 759.0000000000001)]
    assert expected == fraplib.get_em(testfile)
    
    assert [633.0] == fraplib.get_ex(testfile)
    
    assert [650] == fraplib.get_gain(testfile)
    
    assert "Plan-Apochromat 40x/1.3 Oil DIC M27" == fraplib.get_objective(testfile)
     
    assert [0.1] == fraplib.get_power(testfile)
    
    expected = [(128.01983145986762, 128.00010304153278, 37.49985947114415)]
    assert expected == fraplib.get_regions(testfile)
    
    expected = [(5.488461102633269e-09, 2.8517287604581007e-11, 1.0378283869946033e-05)]
    assert expected == fraplib.get_regions(testfile, units="physical")
    
    expected = [(0.005488461102633269, 2.8517287604581008e-05, 10.378283869946033)]
    assert expected == fraplib.get_regions(testfile, units="microns")
    
    expected = {"X-scale": 0.276755273654613, "Y-scale": 0.276755273654613, "Z-scale": 0.0}
    assert expected == fraplib.get_scales(testfile)
    
    expected = {"BLEACH_START": 2827.2208624560003, "BLEACH_STOP": 2833.1649571760004}
    assert expected == fraplib.get_events(testfile)
    


# this code snippet is taken directly from John C. Russell

import numpy as np
import pandas as pd
import xarray as xr

# __all__ = [
#     "check_labels_from_multiindex",
# ]


def check_labels_from_multiindex(
    labels: xr.DataArray, index: pd.MultiIndex, scene_character: str = "S"
) -> xr.DataArray:
    """
    Given a pd.MultiIndex Object which identifies particular cells, return a labels
    array which only shows the cells in the index.
    Parameters
    ----------
    labels: xr.DataArray
        Image like array containg integer labeled regions.
    index: pd.MultiIndex
        MultiIndex object with scene and cell labels to check.
        ***NOTE*** This follows the microutil convention that label values in the index
        are zero indexed - and hence off by one relative to the values in the labels
        array.
    scene_character: str - Default "S"
        Character that refers to the scene dimension of the labels DataArray and the
        scene level of the multiindex.
    """

    level = None
    for i in range(len(index.names)):
        if index.names[i] == scene_character:
            level = i
            break
    if level is None:
        raise KeyError(
            f"Name {scene_character} not found in index with levels {index.names}"
        )

    check_labels = xr.zeros_like(labels)

    for s in range(labels.sizes[scene_character]):
        try:
            # See ***NOTE*** in docstring regarding +1 here
            checks = index.get_loc_level(s, level)[1].values + 1
        except KeyError:
            continue

        mask = xr.DataArray(np.isin(labels.data[s], checks), dims=labels[s].dims)
        check_labels[s] = labels[s].where(mask)

    return check_labels


import numpy as np
from matplotlib import cm
from matplotlib import pyplot as plt
from matplotlib.colors import LinearSegmentedColormap, ListedColormap


def make_colormap(rgb):
    """
    creates a new color map from black to white through the specified RGB color

    Parameters
    ----------
    rgb : list
        [r = #, g = #, b = #] where # is between 0 and 256

    Returns
    -------
    new_cm : mpl.colors.ListedColormap
    """

    r = rgb[0]
    g = rgb[1]
    b = rgb[2]

    N = 256
    vals = np.ones((N, 4))
    vals[:, 0] = np.linspace(0, r / 256, N)
    vals[:, 1] = np.linspace(0, g / 256, N)
    vals[:, 2] = np.linspace(0, b / 256, N)
    ko = ListedColormap(vals)

    N = 256
    vals = np.ones((N, 4))
    vals[:, 0] = np.linspace(r / 256, 1, N)
    vals[:, 1] = np.linspace(g / 256, 1, N)
    vals[:, 2] = np.linspace(b / 256, 1, N)
    ow = ListedColormap(vals)

    top = cm.get_cmap(ko, 128)
    bottom = cm.get_cmap(ow, 128)

    newcolors = np.vstack((top(np.linspace(0, 1, 128)), bottom(np.linspace(0, 1, 128))))

    new_cm = ListedColormap(newcolors)

    return new_cm


def falsecolor(im, cmap=None, *args, **kwargs):
    """
    creates figure and axes objects displaying im with an optionally specified false coloring by cmap

    Parameters
    ----------
    im : np.ndarray
        image to be displayed
    cmap : list or matplotlib.colors.ListedColormap or matplotlib.colors. LinearSegmentedColormap
        [r = #, g = #, b = #] where # is between 0 and 256
    *args : list
        passes arguments to plt.imshow()
    **kwargs : dict
        passes keyword arguments to plt.imshow()

    Returns
    -------
    fig : matplotlib.figure.Figure
    ax : matplotlib.axes._subplots.AxesSubplot
    """

    fig, ax = plt.subplots()

    if cmap is not None:
        if isinstance(cmap, ListedColormap) or isinstance(
            cmap, LinearSegmentedColormap
        ):
            plt.imshow(im, cmap=cmap, *args, **kwargs)
        else:
            cmap = make_colormap(cmap)
            plt.imshow(im, cmap=cmap, *args, **kwargs)
    else:
        plt.imshow(im, *args, **kwargs)

    plt.axis("off")

    return fig, ax



import numpy as np
from scipy import ndimage

def bkgdsubtract(ims, labels):
    """
    Background-subtract a segmented image. For each position and channel, averages all the pixels that were not labeleled as cells during segmentation and subtracts that background value from all pixels in the image.
    
    Parameters
    ----------
    ims : numpy array
        image stack of shape (fovs, channels, Y, X)
    labels: array
        segmentation masks; stack of shape (fovs, Y, X)
    
    Returns
    -------
    bgsub: numpy array
        background-subtracted image stack
    """
    bg = np.zeros_like(ims, dtype='f4')
    for j, fov in enumerate(ims):
        for k, im in enumerate(fov):
            bg[j,k].fill(ndimage.mean(im, labels = labels[j], index = 0))
    bgsub = ims - bg
    return bgsub


try:
    from ._version import version as __version__
except ImportError:
    __version__ = "unknown"
__author__ = "Elizabeth May"
__email__ = "ejmay2012@gmail.com"

from ._setupfunction import _setup
from .loading import load_data, batchread
from .metadatafunctions import (
    version,
    channel_label,
    get_gain,
    get_power,
    get_objective,
    get_em,
    get_ex,
    get_scales,
    get_regions,
    get_dims,
)
from .attachments import get_events, get_timepoints
from .calculations import evaluate, find_biggest_drop
from .circular_mask import create_circular_mask
from .expb2 import fit_expD, fit_expb2
from .falsecolor import falsecolor, make_colormap
from .regions import draw_circle
from .scalebar import scalebar
from .timestamp import timestamp
from .pretty_plot import pretty_plot
from .onestep import analyze
from .read_chromatogram import read_chromatogram
from .bkgdsubtract import bkgdsubtract
from .threshold import threshold
from .filter_labels import check_labels_from_multiindex # JCR

__all__ = [
    "__version__",
    "__author__",
    "__email__",
    "_setup",
    "load_data",
    "batchread",
    "channel_label",
    "evaluate",
    "get_gain",
    "get_power",
    "get_objective",
    "get_em",
    "get_ex",
    "get_scales",
    "get_regions",
    "get_dims",
    "get_events",
    "get_timepoints",
    "create_circular_mask",
    "find_biggest_drop",
    "fit_expD",
    "fit_expb2",
    "falsecolor",
    "make_colormap",
    "draw_circle",
    "scalebar",
    "timestamp",
    "pretty_plot",
    "analyze",
    "bkgdsubtract",
    "threshold",
    "check_labels_from_multiindex",
]



from symfit import variables, Parameter, Fit
import numpy as np
from .metadatafunctions import get_regions

def fit_expD(xdata, ydata, radius, guesses):
    """
    fitting relevant diffusion parameters directly with inverse exponential decay function substituted with FRAP recovery equation
    
    Parameters
    ----------
    xdata : array
    ydata : array
    file : czifile
    guesses : list
        [amount of bleaching, diffusion coefficient, immobile fraction]
    
    Returns
    -------
    xvals : array
        x values for plotting fit
    curve : array
        y values for plotting fit
    fit_result.params : ordered dict
        estimates of model parameters
    fit_result
        output from symfit, including covariance matrix
    """
    
    x, y = variables('x, y')
    
    # _ , _ , radius = np.asarray(get_regions(file, units = 'microns')[0])/1E-6 # in microns
    r = Parameter('r', value = radius, fixed = True) # radius of region in microns
    amt = Parameter('amt', value = guesses[0]) # amount of bleaching
    D = Parameter('D', value = guesses[1]) # diffusion coefficient
    immfr = Parameter('immfr', value = guesses[2]) # immobile fraction

    thalf = (
        0.224 * (r ** 2) * (D ** -1)
    )
    A = (
        amt - immfr
    )
    yoff = (
        1 - immfr
    )
    model = {
        y: yoff - (A * 2**(-x/(0.224 * (r ** 2) * (D ** -1))))
    }

    fit = Fit(model, xdata, ydata)
    fit_result = fit.execute()
    
    xvals = np.linspace(xdata.min(), xdata.max(), num = 100*len(xdata))
    curve = fit.model(x = xvals, **fit_result.params).output[0]
    
    return xvals, curve, fit_result.params, fit_result

def fit_expb2(xdata, ydata, guesses):
    """
    inverse exponential decay function
    
    Parameters
    ----------
    xdata : array
    ydata : array
    guesses : list
        [A, thalf, yoff]
    
    Returns
    -------
    xvals : array
        x values for plotting fit
    curve : array
        y values for plotting fit
    fit_result.params : ordered dict
        estimates of model parameters
    fit_result
        output from symfit, including covariance matrix
    """
    
    A, thalf, yoff = guesses # unpack
    
    x, y = variables('x, y')
    A = Parameter('A', value = A)
    thalf = Parameter('thalf', value = thalf)
    yoff = Parameter('yoff', value = yoff)
    
    model = { y: yoff - (A * 2**(-x/thalf)) }
    
    fit = Fit(model, xdata, ydata)
    fit_result = fit.execute()
    
    xvals = np.linspace(xdata.min(), xdata.max(), num = 100*len(xdata))
    curve = fit.model(x = xvals, **fit_result.params).output[0]
    
    return xvals, curve, fit_result.params, fit_result


from matplotlib import pyplot as plt
from mpl_interactions import hyperslicer
from .metadatafunctions import get_regions, get_scales
from .regions import draw_circle
from .scalebar import scalebar


def pretty_plot(images, xdata, ydata, fitx, fity, experiment, name = None):
    """
    Plot the results of a FRAP experiment nicely
    
    Parameters
    ----------
    images : array
    xdata : array
        timepoints adjusted to t_0 = t_bleach
    ydata : array
        normalized fluorescence data for each timepoint
    fitx : array
        post-bleach timepoints used for fitting
    fity : array
        output curve datapoints from fitting
    experiment : CziFile
        original data file, for metadata
    name : str
        experiment name for figure
    """
    
    fig, axs = plt.subplots(1, 2, dpi=120, figsize=(6.5, 3), layout="constrained")

    left, right = axs
    cyan = (0,1,1)
    dcyan = (0,0.8, 0.8)
    gray = 3*(0.7,)

    for cnt, series in enumerate(ydata):
        right.scatter(xdata[xdata >= 0], series[xdata >= 0], label="data", color = cyan)
        right.plot(fitx[cnt], fity[cnt], label="fit", color = dcyan)
        right.scatter(xdata, series, label="normalization data", color=gray, zorder = 0)
        
    right.axhline(
        y=1, color='k', zorder=-1, label="normalization factor"
    )
    left.set_title("FRAP", color="gray")
    right.set_xlabel("time (seconds)")
    right.set_ylabel("normalized fluorescence")
    # right.legend(loc = 'lower right', frameon = False)

    limits = (images[xdata >= 0, ...][0].min(), images[xdata >= 0, ...][0].max())
    ctrls = hyperslicer(
        images,
        ax=left,
        names=["frame"],
        play_buttons=True,
        cmap="gray",
        vmin=limits[0],
        vmax=limits[1],
    )
    left.set_axis_off()
    circle = get_regions(experiment)
    for region in circle:
        draw_circle(region, ax=left, color=cyan, linewidth=1)
    scales = get_scales(experiment)
    scalebar(20, images, scales, color="w", label_on=True, ax=left)
    if name is not None:
        left.set_title(name, color="gray")

    plt.show()
    
    return fig, axs


from .calculations import evaluate
from .metadatafunctions import get_regions
from .expb2 import fit_expD
from .pretty_plot import pretty_plot

def analyze(czif, name = None):
    """
    One-step analysis of a FRAP experiment.
    
    Parameters
    ----------
    czif : CziFile
    
    Returns
    -------
    fig : matplotlib figure object
    axs : matplotlib axes object
    p : list
        contains ordered dicts of fit params for each region in the experiment
    """
    
    # step 1: get the images
    images = czif.asarray().squeeze()

    # step 2: do all the calculations
    t, n, _, _, _, _, _, _, _ = evaluate(images, czif)

    # step 3: fit post-bleach data to exponential decay with base 2
    xdata = t[t >= 0]
    # ydata = n[t >= 0]
    roi = get_regions(czif, units = 'microns')
    curves = []
    p = []
    fit_results = []
    fitxval = []
    for cnt, region in enumerate(roi):
        radius = region[-1]
        ydata = n[cnt][t >= 0]
        guesses = (0.5, 1, 0.1)
        fitxvals, curve, params, fit_result = fit_expD(xdata, ydata, radius, guesses)
        curves.append(curve)
        p.append(params)
        fit_results.append(fit_result)
        fitxval.append(fitxvals)

    # step 4: plot
    fig, axs = pretty_plot(
        images, 
        xdata = t, ydata = n, 
        fitx = fitxval, fity = curves, 
        experiment = czif,
        name = name
    )
    
    return fig, axs, p


from matplotlib import pyplot as plt


def timestamp(frame, timestamps, fsize=6, t_units='sec', ax=None, color=None):
    """
    adds a time stamp to an image

    Parameters
    ----------
    frame : index
    timestamps : array
    fsize : int or float
    t_units : str
        'sec' or 'min'
    ax: matplotlib.axes._subplots.AxesSubplot
    """

    if ax is None:
        ax = plt.gca()

    if color is None:
        color = next(ax._get_lines.prop_cycler)['color']

    t = round(timestamps[frame])

    if t_units == 'sec':
        time = str(round(t))
        t_str = time + ' sec'
    elif t_units == 'min':
        time = str(round(t / 60))
        t_str = time + ' min'

    ax.annotate(t_str, (0.1, 0.1), xycoords='axes fraction', color=color, size=fsize)



import czifile

def get_events(file):
    """
    Grabs experiment 'event' descriptions and affiliated timestamps.
    
    Parameters
    ----------
    file : CziFile
        original data file
    
    Returns
    -------
    eventlist : dict
        { description : timestamp }
    """
    
    eventlist = {}
    
    for atch in file.attachments():
        if 'Event' in atch.attachment_entry.name:
            evl = atch.data()
            for ind in range(len(evl)):
                event = str(evl[ind])
                desc = event.split()[0]
                tval = float(event.split('@')[1].split()[0])
                eventlist[desc] = tval
    
    return eventlist

def get_timepoints(file):
    """
    Grabs experiment frame timepoints.
    
    Parameters
    ----------
    file : CziFile
        original data file
    
    Returns
    -------
    t : array
        array of timepoints as floats
    traw : array
        non-zeroed array of timepoints as floats, for comparison with e.g. event timestamps
    """
    
    for atch in file.attachments():
        if 'TimeStamps' in atch.attachment_entry.name:
            traw = atch.data()
    t = traw - traw.min()
    
    return t, traw


from pathlib import Path
import os
import czifile


def load_data(path):
    """
    loads the file and the metadata, including frame-specific metadata

    Parameters
    ----------
    path : str
        path to the file

    Returns
    -------
    file : CziFile
        original data file in CziFile format
    subblocks : dict
        { images : array, immd : str }
        # check if subblocks["images"] is really an array
        # check if subblocks["immd"] is really str
    """

    file = czifile.CziFile(str(Path(path).resolve()))

    subblocks = {
        "images": [
            segment.data()
            for segment in file.segments()
            if isinstance(segment, czifile.SubBlockSegment)
        ],
        "immd": [ # image-specific metadata
            segment.metadata()
            for segment in file.segments()
            if isinstance(segment, czifile.SubBlockSegment)
            # and segment.metadata() is not None
        ],
    }
    
    return file, subblocks

def batchread(directory):
    """
    Reads .czi files in directory into a data holder [holder].
    
    Parameters
    ----------
    directory : str
        path to directory containing files of interest
    
    Returns
    -------
    holder : dict
        keys correspond to a file name; values correspond to the data (type = CziFile)
    pathholder : dict
        keys correspond to a file name; values correspond to a path to the file (type = str)
    """
    
    # directory = input_dir
    
    extension = '.czi'
    
    holder = {}
    pathholder = {}
    
    for item in os.listdir(directory):
        
        f = os.path.join(directory, item)
        
        if os.path.isfile(f) and item[-4:] == extension:
            
            # print(item)
            file, _ = load_data(f)
            holder[item] = file
            pathholder[item] = f
    
    return holder, pathholder


import matplotlib.patches as patches
from matplotlib import pyplot as plt


def scalebar(length, im, im_scales, ax=None, color=None, label_on = False, **kwargs):
    """
    adds a scale bar to a plot

    Parameters
    ----------
    length : number
        desired scale bar length in image pixel units (typically microns)
    im : array
        2- or 3-D image
    im_scales : list
        [x µm/px, y µm/px, z µm/px]
    dims : list
        [x px, y px]
    ax : matplotlib.axes._subplots.AxesSubplot
        axes object on which to put the scale bar
    color : str
        color of the scalebar
    label_on : bool
        text label under the scale bar on/off
    **kwargs : dict
        keyword arguments to pass to matplotlib.patches.Rectangle()
    
    Returns
    -------
    """

    pxsize = im_scales[0] # x µm/px
    xpx = im.shape[-1]
    ypx = im.shape[-2]

    if ax is None:
        ax = plt.gca()

    if color is None:
        color = next(ax._get_lines.prop_cycler)['color']

    px_length = length / pxsize
    fractional = px_length / xpx

    # build a rectangle in axes coords
    left, width = 0.9 - fractional, fractional
    bottom, height = 0.1, 0.02
    right = left + width
    top = bottom + height

    # axes coordinates: (0, 0) is bottom left and (1, 1) is upper right
    p = patches.Rectangle(
        (left, bottom),
        width,
        height,
        facecolor=color,
        edgecolor=None,
        transform=ax.transAxes,
    )

    ax.add_patch(p)
    
    if label_on:
        ax.text(left + fractional/2, bottom - bottom/2, r'$'+str(length)+'\ µm$',
                horizontalalignment='center',
                verticalalignment='center',
                transform=ax.transAxes,
                color = color
               )


from matplotlib import pyplot as plt
import matplotlib.patches as patches

def draw_circle(roi, ax = None, color = None, **kwargs):
    """
    adds a circle corresponding to roi onto an image

    Parameters
    ----------
    roi : tuple
        (centerX, centerY, radius)
    ax : matplotlib.axes._subplots.AxesSubplot
        axes object on which to put the scale bar
    color : str
        color of the circle
    **kwargs : dict
        keyword arguments to pass to matplotlib.patches.Circle()

    Returns
    -------
    """
    
    if ax is None:
        ax = plt.gca()
    
    if color is None:
        color = next(ax._get_lines.prop_cycler)['color']
    
    x = roi[0]
    y = roi[1]
    r = roi[2]

    c = patches.Circle((x, y), r, facecolor='none', edgecolor=color, **kwargs)
    
    ax.add_patch(c)


import numpy as np


def create_circular_mask(image, circle_geom):
    """
    create a circular mask

    Parameters
    ----------
    image : array
    circle_geom : tuple
        (centerx, centery, radius)

    Returns
    -------
    mask : bool
    """
    w = image.squeeze().shape[-1]
    h = image.squeeze().shape[-2]

    center = (circle_geom[0], circle_geom[1])
    radius = circle_geom[2]

    if center is None:  # use the middle of the image
        center = (int(w / 2), int(h / 2))
    if radius is None:  # use the smallest distance between the center and image walls
        radius = min(center[0], center[1], w - center[0], h - center[1])

    Y, X = np.ogrid[:h, :w]
    dist_from_center = np.sqrt((X - center[0]) ** 2 + (Y - center[1]) ** 2)

    mask = dist_from_center <= radius
    return mask


import numpy as np
from .attachments import get_events, get_timepoints
from .metadatafunctions import get_regions
from .circular_mask import create_circular_mask

def evaluate(imageseries, datafile, background = None, mask = None, bleachequiv = None):
    """
    Raw data processing for FRAP timeseries.
    
    Workflow:
        background subtraction
        average pixels in bleached region at each timepoint (m)
        average signal for timepoints before bleach (normalization factor)
        normalize signal at each time point relative to normalization factor (n)
        
    Parameters
    ----------
    imageseries : array
    datafile : CziFile
    background : number
        value to use for background subtraction
    mask : bool
        same shape as images; True for pixels in bleach region
    bleachequiv : int
        number of frames to average for normalization when there is no bleach
    
    Returns
    -------
    tpost : array
        time points after bleach
    npost : array
        normalized signal for data points after bleach
    t : array
        all time points relative to t_0 which corresponds to t_bleach
    n : array
        all normalized data points
    traw : array
        all time points as extracted from the metadata
    bleach : float
        time point at which bleach occurred
    normfactor: float
        value to use for data normalization
    m : array
        average signal in bleach region
    bsim : array
        background subtracted image series
    mask : bool
        same shape as images; True for pixels in bleach region
    background : number
        value to use for background subtraction
    pre : bool
        same shape as t; True corresponding to t before bleach
    post : bool
        same shape as t; True corresponding to t after bleach
    """
    
    if background is None:
        background = 0
    
    # time and events
    if bleachequiv is None:
        bleach = get_events(datafile)['BLEACH_START']
    else:
        bleach = 0
        nnorm = bleachequiv # number of frames to average for normalization when there is no bleach
    
    t, traw = get_timepoints(datafile)
    pre = traw < bleach # bool
    post = traw > bleach # bool
    
    tpost = t[post]
    tpre = t[pre] - tpost.min()
    t -= tpost.min()
    tpost -= tpost.min()
    
    # background subtraction
    bsim = imageseries - background # background-subtracted image
    
    
    if mask is None:
        circle = get_regions(datafile)
        m = []
        normfactor = []
        n = []
        npost = []
        for counter, region in enumerate(circle):
            if isinstance(region, tuple):
                mask = create_circular_mask(imageseries, region)

                # averaging pixels in bleach region
                m.append(bsim[...,mask].mean(axis = -1)) # list of arrays; mean of px in each bleached region for each timepoint

                # normalization
                if bleachequiv is None:
                    normfactor.append(m[counter][pre].mean())
                else:
                    normfactor.append(m[counter][0:bleachequiv-1].mean())
                n.append(m[counter]/normfactor[counter]) # all the datapoints
                npost.append(m[counter][post]/normfactor[counter]) # just the datapoints after the bleach
    
    return t, n, traw, bleach, normfactor, m, bsim, mask, background


def find_biggest_drop(data):
    """
    Returns the (right) array position of the biggest difference between two adjacent values in an array.
    
    Parameters
    ----------
    data : array
    
    Returns
    -------
    index : int
    """
    
    drops = data[:-1] - data[1:]
    biggest = drops.min()
    index = np.flatnonzero(biggest == drops)[0]
    return index


import pandas as pd
from matplotlib import pyplot as plt
import numpy as np
import xarray as xr

def threshold(avgs, channels, channel_control_samples, Q = 0.995):
    """
    Compute channel-wise thresholds.
    
    Parameters
    ----------
    avgs : pd Series
    channels: list
    channel_control_samples: list
    Q: float
    
    Returns
    -------
    threshold: pd Series
    """
    
    threshold = pd.Series(dtype='f4')
    fig, axs = plt.subplots(1,2,layout = 'constrained', figsize = (8, 3.5), sharey = True)

    chs = channels[:-1]
    
    chclr = {'Alexa Fluor 546' : 'm', 'Cy5' : 'm', 'EGFP' : 'c', 'mCherry' : 'r'}
    
    clrs = [chclr[ch] for ch in chs]
    
    sels = channel_control_samples
    # sels = [
    #     list(ximages.S.values[6:]), # samples with no AF546 fluorescence
    #     [s for s in ximages.S.values if 'noFP' in s] # samples with no EGFP fluorescence
    # ]

    for a, channel, sel, c in zip(np.arange(len(chs)), chs,sels,clrs):

        rangemin = avgs.loc[pd.IndexSlice[sel,channel]].min()
        rangemax = avgs.loc[pd.IndexSlice[sel,channel]].max()

        for s in sel:
            axs[a].hist(avgs.loc[pd.IndexSlice[s,channel]],
                        bins = 100,
                        range=(rangemin, rangemax),
                        alpha = 1/(3*len(sel)),
                        label = s,
                        color = c,)
                       # density = True)

        threshold[channel] = avgs.loc[pd.IndexSlice[sel,channel]].quantile(Q)

        axs[a].axvline(threshold[channel], color = 'k')
        axs[a].semilogy()
        # ax[a].legend()
        axs[a].set_title(f'{channel}')
        axs[a].set_xlabel(f'Mean cell intensity in {channel} channel')
        if a ==0:
            axs[a].set_ylabel('Number of cells\n(log scale)')
    plt.suptitle('Negative control histograms for thresholding')
    # plt.savefig(o+'thresholding.svg', dpi = 300)
    plt.show()
    
    return threshold